import click
import docker
import yaml
import time
from pathlib import Path
from typing import Dict, Any, List, Tuple
from concurrent.futures import ThreadPoolExecutor, as_completed
import sys

from .load_balancer import LoadBalancerManager

class DockerManager:
    def __init__(self):
        try:
            self.client = docker.from_env()
        except docker.errors.DockerException as e:
            click.secho(f"Failed to initialize Docker client: {e}", fg='red', err=True)
            sys.exit(1)
        self.config = {}
        self.services = {}
        self.load_balancers = {}
        self._containers = set()  # Track containers for proper cleanup
        try:
            self.lb_manager = LoadBalancerManager(self.client)
        except Exception as e:
            click.secho(f"Failed to initialize load balancer: {e}", fg='red', err=True)
            sys.exit(1)

    def load_config(self, config_file: str) -> None:
        """Load configuration from orca.yml file"""
        try:
            with open(config_file, 'r') as f:
                self.config = yaml.safe_load(f) or {}
                if not isinstance(self.config, dict):
                    raise ValueError("Invalid configuration format")
                self.services = self.config.get('services', {})
                self.load_balancers = self.config.get('load_balancers', {})
        except (yaml.YAMLError, IOError) as e:
            click.secho(f"Failed to load config file {config_file}: {e}", fg='red', err=True)
            sys.exit(1)
        except ValueError as e:
            click.secho(f"Invalid configuration in {config_file}: {e}", fg='red', err=True)
            sys.exit(1)

    def _start_container(self, name: str, instance_name: str, config: Dict[str, Any], network_name: str, rebuild: bool) -> Tuple[str, str, str, bool]:
        if not isinstance(config, dict):
            click.secho(f"Invalid configuration for service {name}", fg='red', err=True)
            return name, instance_name, "Invalid configuration", False
        """Start a single container and return its status
        
        Returns:
            Tuple of (service_name, instance_name, container_id, success)
        """
        try:
            image = config.get('image')
            expose_ports = config.get('expose', [])
            environment = config.get('environment', {})
            volumes = config.get('volumes', [])

            # Set up port exposures (no host binding)
            port_bindings = {}
            for port in expose_ports:
                port_bindings[f"{port}/tcp"] = None

            # Convert volumes to docker format
            volume_bindings = {}
            for volume in volumes:
                host_path, container_path = volume.split(':')
                volume_bindings[str(Path(host_path).absolute())] = {
                    'bind': container_path,
                    'mode': 'rw'
                }

            # Pull/rebuild image if requested
            if rebuild:
                self.client.images.pull(image)

            container = self.client.containers.run(
                image=image,
                name=instance_name,
                detach=True,
                ports=port_bindings,
                environment=environment,
                volumes=volume_bindings,
                network=network_name
            )
            return name, instance_name, container.short_id, True
        except docker.errors.APIError as e:
            click.echo(f"Error starting {instance_name}: {str(e)}", err=True)
            return name, instance_name, str(e), False

    def up(self, service_name: str = None, rebuild: bool = False) -> None:
        """Start services defined in the configuration with scaling support
        
        Args:
            service_name: Optional name of specific service to start
            rebuild: If True, force rebuild/pull of images before starting
        """
        services_to_start = {service_name: self.services[service_name]} if service_name else self.services

        # Create a custom network if it doesn't exist
        network_name = 'orca_network'
        try:
            network = self.client.networks.get(network_name)
        except docker.errors.NotFound:
            network = self.client.networks.create(
                network_name,
                driver='bridge',
                attachable=True
            )

        # Get total number of tasks (services + network)
        total_tasks = len(services_to_start) + 1  # +1 for network
        click.secho(f"[+] Running {total_tasks}/{total_tasks}", fg='bright_black')

        # Create network first
        network_start = time.time()
        click.secho(f" ✔ Network orca_default    Created", nl=False, fg='green')
        click.secho(f"{time.time() - network_start:>50.1f}s", fg='bright_black')

        # Prepare tasks
        tasks = []
        for name, config in services_to_start.items():
            scale = config.get('scale', 1)
            start_time = time.time()
            
            for instance in range(scale):
                instance_name = f"{name}_{instance + 1}" if scale > 1 else name
                tasks.append((name, instance_name, config))

        # Start containers in parallel
        with ThreadPoolExecutor(max_workers=min(32, len(tasks))) as executor:
            futures = []
            for name, instance_name, config in tasks:
                future = executor.submit(self._start_container, name, instance_name, config, network_name, rebuild)
                futures.append(future)
            
            # Process results as they complete
            service_status = {}
            for future in as_completed(futures):
                name, instance_name, container_id, success = future.result()
                if name not in service_status:
                    service_status[name] = {'success': True, 'instances': []}
                
                service_status[name]['success'] &= success
                service_status[name]['instances'].append((instance_name, container_id))
                
                # If all instances for a service are complete, show the status
                if len(service_status[name]['instances']) == services_to_start[name].get('scale', 1):
                    if service_status[name]['success']:
                        container_name = f"orca-{name}-1"
                        click.secho(f" ✔ Container {container_name:<12} Created", nl=False, fg='green')
                        click.secho(f"{time.time() - start_time:>50.1f}s", fg='bright_black')
                        instances = service_status[name]['instances']
                        if len(instances) > 1:
                            for inst_name, cont_id in instances:
                                click.echo(f"  {inst_name}: {cont_id}")
                    else:
                        click.echo('\033[F\033[0G', nl=False)
                        click.secho(f"Creating {name}...", nl=False, fg='bright_black')
                        click.secho(" Failed", nl=True, fg='red')
                        click.echo(f"Error: Failed to create one or more instances of {name}")
        
        # Wait for services to be ready before starting load balancers
        if not service_name and self.load_balancers:
            def wait_for_service(service_name: str, scale: int) -> bool:
                for i in range(scale):
                    instance_name = f"{service_name}_{i + 1}" if scale > 1 else service_name
                    max_retries = 10
                    retry_count = 0
                    while retry_count < max_retries:
                        try:
                            container = self.client.containers.get(instance_name)
                            if container.status == 'running':
                                break
                        except:
                            pass
                        retry_count += 1
                        time.sleep(1)
                    if retry_count >= max_retries:
                        return False
                return True

            def create_lb(lb_name: str, lb_config: Dict) -> Tuple[str, bool]:
                try:
                    # Wait for all required services
                    for service_config in lb_config.get('services', []):
                        service_name = service_config['name']
                        service_scale = self.services[service_name].get('scale', 1)
                        if not wait_for_service(service_name, service_scale):
                            return lb_name, False

                    self.lb_manager.create_load_balancer(lb_name, lb_config, self.services)
                    return lb_name, True
                except Exception as e:
                    click.echo(f"Error creating load balancer {lb_name}: {str(e)}", err=True)
                    return lb_name, False

            # Start load balancers in parallel
            with ThreadPoolExecutor(max_workers=min(8, len(self.load_balancers))) as executor:
                futures = []
                for lb_name, lb_config in self.load_balancers.items():
                    click.secho(f"Creating load balancer {lb_name}...", nl=False, fg='bright_black')
                    sys.stdout.flush()
                    future = executor.submit(create_lb, lb_name, lb_config)
                    futures.append((lb_name, future))

                # Process results as they complete
                for lb_name, future in futures:
                    _, success = future.result()
                    if success:
                        click.secho(" Done\n", fg='green')
                    else:
                        click.secho(" Failed\n", fg='red')

    def _stop_container(self, name: str, instance_name: str) -> Tuple[str, str, str, bool]:
        """Stop and remove a single container
        
        Returns:
            Tuple of (service_name, instance_name, container_id, success)
        """
        try:
            container = self.client.containers.get(instance_name)
            container.stop()
            container.remove()
            return name, instance_name, container.short_id, True
        except docker.errors.NotFound:
            click.echo(f"Container {instance_name} not found")
            return name, instance_name, "not found", False
        except docker.errors.APIError as e:
            click.echo(f"Error stopping {instance_name}: {str(e)}", err=True)
            return name, instance_name, str(e), False

    def down(self, service_name: str = None) -> None:
        """Stop and remove containers including scaled instances and load balancers"""
        # First stop load balancers if we're stopping all services
        if not service_name and self.load_balancers:
            with ThreadPoolExecutor(max_workers=min(8, len(self.load_balancers))) as executor:
                futures = []
                for lb_name in self.load_balancers:
                    click.secho(f"Removing load balancer {lb_name}...", nl=False, fg='bright_black')
                    sys.stdout.flush()
                    future = executor.submit(self.lb_manager.remove_load_balancer, lb_name)
                    futures.append((lb_name, future))
                
                # Process results as they complete
                for lb_name, future in futures:
                    try:
                        future.result()
                        click.secho(" Done\n", fg='green')
                    except Exception as e:
                        click.secho(" Failed\n", fg='red')
                        click.echo(f"Error removing load balancer {lb_name}: {str(e)}", err=True)

        # Then stop services
        services_to_stop = [service_name] if service_name else self.services.keys()
        
        # Get total number of tasks
        total_tasks = len(services_to_stop)
        click.secho(f"[+] Running {total_tasks}/{total_tasks}", fg='bright_black')

        # Prepare tasks
        tasks = []
        for name in services_to_stop:
            scale = self.services[name].get('scale', 1) if name in self.services else 1
            start_time = time.time()
            
            for instance in range(scale):
                instance_name = f"{name}_{instance + 1}" if scale > 1 else name
                tasks.append((name, instance_name))
        
        # Stop containers in parallel
        with ThreadPoolExecutor(max_workers=min(32, len(tasks))) as executor:
            futures = []
            for name, instance_name in tasks:
                future = executor.submit(self._stop_container, name, instance_name)
                futures.append(future)
            
            # Process results as they complete
            service_status = {}
            for future in as_completed(futures):
                name, instance_name, container_id, success = future.result()
                if name not in service_status:
                    service_status[name] = {'success': True, 'instances': []}
                
                service_status[name]['success'] &= success
                service_status[name]['instances'].append((instance_name, container_id))
                
                # If all instances for a service are complete, show the status
                if len(service_status[name]['instances']) == (self.services[name].get('scale', 1) if name in self.services else 1):
                    if service_status[name]['success']:
                        container_name = f"orca-{name}-1"
                        click.secho(f" ✔ Container {container_name:<12} Removed", nl=False, fg='green')
                        click.secho(f"{time.time() - start_time:>50.1f}s", fg='bright_black')
                        instances = service_status[name]['instances']
                        if len(instances) > 1:
                            for inst_name, cont_id in instances:
                                click.echo(f"  {inst_name}: {cont_id}")
                    else:
                        container_name = f"orca-{name}-1"
                        click.secho(f" ✘ Container {container_name:<12} Failed", nl=False, fg='red')
                        click.secho(f"{time.time() - start_time:>50.1f}s", fg='bright_black')
                        click.echo(f"Error: Failed to stop one or more instances of {name}")

    def ps(self) -> None:
        """List running containers with scale information and load balancers"""
        containers = self.client.containers.list()
        if not containers:
            click.echo("No running containers")
            return

        # Group containers by service name
        service_containers = {}
        lb_containers = []
        for container in containers:
            if container.name in self.load_balancers:
                lb_containers.append(container)
            else:
                service_name = container.name.rsplit('_', 1)[0] if '_' in container.name else container.name
                if service_name not in service_containers:
                    service_containers[service_name] = []
                service_containers[service_name].append(container)

        # Print services
        if service_containers:
            click.echo("\nSERVICES:")
            click.echo("SERVICE\t\tSCALE\tCONTAINER ID\tNAME\t\tSTATUS\t\tPORTS")
            for service_name, containers in service_containers.items():
                scale = len(containers)
                for container in containers:
                    # Handle port mapping format from Docker API
                    port_mappings = container.attrs['NetworkSettings']['Ports']
                    port_list = []
                    if port_mappings:
                        for container_port, host_bindings in port_mappings.items():
                            if host_bindings:  # Check if port is actually mapped
                                for binding in host_bindings:
                                    host_port = binding['HostPort']
                                    # Remove the protocol suffix if present (e.g., '80/tcp' -> '80')
                                    container_port = container_port.split('/')[0]
                                    port_list.append(f"{host_port}->{container_port}")
                    ports = ', '.join(port_list)
                    click.echo(f"{service_name}\t\t{scale}\t{container.short_id}\t{container.name}\t{container.status}\t{ports}")
        
        # Print load balancers
        if lb_containers:
            click.echo("\nLOAD BALANCERS:")
            click.echo("NAME\t\tCONTAINER ID\tSTATUS\t\tPORTS")
            for container in lb_containers:
                port_mappings = container.attrs['NetworkSettings']['Ports']
                port_list = []
                if port_mappings:
                    for container_port, host_bindings in port_mappings.items():
                        if host_bindings:
                            for binding in host_bindings:
                                host_port = binding['HostPort']
                                container_port = container_port.split('/')[0]
                                port_list.append(f"{host_port}->{container_port}")
                ports = ', '.join(port_list)
                click.echo(f"{container.name}\t\t{container.short_id}\t{container.status}\t{ports}")

@click.group()
def main():
    """Orca - A Docker Compose alternative"""
    pass

@main.command()
@click.option('--file', '-f', default='orca.yml', help='Path to Orca configuration file')
@click.option('--rebuild', is_flag=True, help='Force rebuild/pull of images before starting')
@click.argument('service', required=False)
def up(file: str, service: str = None, rebuild: bool = False):
    """Start services"""
    manager = DockerManager()
    manager.load_config(file)
    manager.up(service, rebuild)

@main.command()
@click.argument('service', required=False)
def down(service: str):
    """Stop and remove services"""
    manager = DockerManager()
    manager.load_config('orca.yml')
    manager.down(service)

@main.command()
def ps():
    """List running containers"""
    manager = DockerManager()
    manager.ps()

if __name__ == '__main__':
    main()
